/*

Name: XCheckVoid

Version: 1.0
 
Description: execute an XML benchmark and saves both query processing times and answers. Methods parseInputDocument (to parse a XML document), executeInputQuery (to
execute a query), and serializeResults (to serialize the query result) do 
nothing and need to be overridden.

Author: Copyright (C) 2007 Massimo Franceschet (massimo.franceschet@dimi.uniud.it) - Dipartimento di Matematica e Informatica, Universita' degli Studi di Udine, Via delle Scienze, 206 - 33100 Udine (Italy)

License: GNU General Public License (GPL)    
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


import java.io.*;
import org.w3c.dom.*;
import org.w3c.dom.ls.*;
import javax.xml.*;
import javax.xml.parsers.*;
import javax.xml.xpath.*;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXException;


/******************************************************************************
Execute an XML benchmark and saves both query processing times and answers. 
Methods parseInputDocument (to parse a XML document), executeInputQuery (to
execute a query), and serializeResults (to serialize the query result) do 
nothing and should be overridden.
******************************************************************************/
public abstract class XCheckVoid {
  
  /****************************************************************
  Variables that are used in main and in at least another procedure
  *****************************************************************/
  public int numQueries; 
  public String[] queryID; 
  public String[] queryDesc; 
  public String[] queryType;
  public String[] querySyntax;
  public String[] queryFileName;
  public String[] queryTargetID;
  public String[] queryTargetDesc;
  public String[] queryTargetFile;
  public String engineID;
  public String engineName;
  public String engineVersion;
  public String engineDesc;
  public String engineHomepage;
  public String engineExecutable;
  public String engineBeforeRun;
  public String engineAfterRun;
  public long[] timeVectorParse; 
  public long[] timeVectorExecute; 
  public long[] timeVectorSerialize; 
  public long[] timeVectorOverall; 
  public boolean[] errorDocument;
  public String[] errorMessageDocument; 
  public boolean[] errorQuery;
  public String[] errorMessageQuery; 
  public Boolean errors; 
  public String xmlOption;
  public String plotOption;
  public String htmlOption;
  public String timesOption;
  public String answersOption;
  public String gnuplotPath;
  public String parseOption;
  public String executeOption;
  public String serializeOption;
  public String overallOption;
  public String checkOutcome;
  public int runs;
  public final String SEP = File.separator;
  public String xcheckHome = "";
  public String experimentFolder; 
  public String experimentName;
  public String experimentType;
  public String experimentDesc;
  public String adapterFileName;
  public String fullAdapterFileName;
  public File folder = null;

  
/******************************************************************************
Execute an XML benchmark and saves both query processing times and answers. 
******************************************************************************/
  public void check(String[] args) {
    
  if (args.length != 2) {
    System.out.println("Usage: java XCheck engine experiment");
    System.out.println("where engine is the name of the engine's adapter " + 
        "and experiment is the name of the experiment folder");
    System.exit(1);
  }
    
    /*************************************** 
    Check for existence of experiment folder   
    ****************************************/
    adapterFileName = args[0] + ".xml";
    fullAdapterFileName = xcheckHome + "adapters" + SEP + adapterFileName;
    folder = new File(fullAdapterFileName);
    if (!folder.exists()) {
      printErrorMessage("An adapter with name " + args[0] + 
          " does not exist in folder adapters. I stop here.");
      System.exit(1);
    }
        
    folder = new File(xcheckHome + "experiments" + SEP + args[1]);
    if (folder.exists()) {
      experimentFolder = "experiments" + SEP + args[1] + SEP;
    } else {
      printErrorMessage("A folder with name " + args[1] + 
          " does not exist in folder experiments. I stop here.");
      System.exit(1);
      
    }
        
    /********************************** 
    Parse XCheck configuration document   
    ***********************************/
    System.out.println();
    System.out.println("Reading XCheck configuration document...");
    DocumentBuilderFactory factoryConfig = DocumentBuilderFactory.newInstance();
    /* This is to validate the document wrt the DTD specified in the 
    Document Type Declaration */
    factoryConfig.setValidating(true);
    DocumentBuilder parserConfig = null;
    try {
      parserConfig = factoryConfig.newDocumentBuilder();
    } catch (ParserConfigurationException pce) {pce.printStackTrace();}
    parserConfig.setErrorHandler(new StrictErrorHandler());
    File configFile = new File(xcheckHome + "config.xml");
    Document config = null;
    try {
      config = parserConfig.parse(configFile);
    } catch (IOException ioe) {
      printErrorMessage(
          "There was a problem while parsing configuration document. I stop here.",
                        ioe.toString());
      System.exit(1);
    } catch (SAXException se) {
      printErrorMessage(
          "There was a problem while parsing configuration document. I stop here",
                        se.toString());
      System.exit(1);
    } 
        
    // Retrieve options from the configuration document
    XPath xpath = XPathFactory.newInstance().newXPath();
    try {
      xmlOption = xpath.evaluate("/config/xml/@value", config); 
    } catch (XPathExpressionException xee) {xee.printStackTrace();}

    try {
      plotOption = xpath.evaluate("/config/plot/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      htmlOption = xpath.evaluate("/config/html/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      timesOption = xpath.evaluate("/config/times/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
      
    try {
      answersOption = xpath.evaluate("/config/answers/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {    
      parseOption = xpath.evaluate("/config/parse/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {   
      executeOption = xpath.evaluate("/config/execute/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      serializeOption = xpath.evaluate("/config/serialize/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      overallOption = xpath.evaluate("/config/overall/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      checkOutcome = xpath.evaluate("/config/checkOutcome/@value", config);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      gnuplotPath = xpath.evaluate("/config/gnuplotPath", config).trim();
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    if (!gnuplotPath.equals("")) {
      gnuplotPath = gnuplotPath + SEP;
    }
      
    try {
      runs = Integer.parseInt(xpath.evaluate("/config/runs", config));
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    if (runs < 1) {
      printErrorMessage(
          "The number of runs in the configuration file " +
          "must be an integer bigger than zero. I stop here.");
      System.exit(1);
    }
    
    /********************* 
    Parse adapter document   
    **********************/
    System.out.println();
    System.out.println("Reading adapter document...");
    DocumentBuilderFactory factoryAdapter = DocumentBuilderFactory.newInstance();
    factoryAdapter.setValidating(true);
    DocumentBuilder parserAdapter = null;
    try {
      parserAdapter = factoryAdapter.newDocumentBuilder();
    } catch (ParserConfigurationException pce) {pce.printStackTrace();}
    parserAdapter.setErrorHandler(new StrictErrorHandler());
    File adapterFile = new File(fullAdapterFileName);
    Document adapter = null;
    try {
      adapter = parserAdapter.parse(adapterFile);
    } catch (IOException ioe) {
      printErrorMessage(
          "There was a problem while parsing the adapter document. I stop here",
                       ioe.toString());
      System.exit(1);
    } catch (SAXException se) {
      printErrorMessage(
          "There was a problem while parsing the adapter document. I stop here",
                        se.toString());
      System.exit(1);
    } 
    
    // Query the adapter document
    xpath = XPathFactory.newInstance().newXPath();
    try {
      // Retrieve engine ID
      String search = "/engine/@id";
      engineID = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      // Retrieve engine name
      String search = "/engine/name";
      engineName = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    
    try {
      // Retrieve engine version
      String search = "/engine/version";
      engineVersion = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}

    try {
      // Retrieve engine description
      String search = "/engine/description";
      engineDesc = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}

    try {
      // Retrieve engine homepage
      String search = "/engine/homepage";
      engineHomepage = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      // Retrieve engine executable
      String search = "/engine/executable";
      engineExecutable = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    try {
      // Retrieve engine before-run command
      String search = "/engine/beforeRun";
      engineBeforeRun = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}

    try {
      // Retrieve engine after-run command
      String search = "/engine/afterRun";
      engineAfterRun = xpath.evaluate(search, adapter);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}

    
    /************************ 
    Parse experiment document   
    *************************/
    System.out.println();
    System.out.println("Reading the experiment...");
    DocumentBuilderFactory factoryExperiment = DocumentBuilderFactory.newInstance();
    factoryExperiment.setValidating(true);
    DocumentBuilder parserExperiment = null;
    try {
      parserExperiment = factoryExperiment.newDocumentBuilder();
    } catch (ParserConfigurationException pce) {pce.printStackTrace();}
    parserExperiment.setErrorHandler(new StrictErrorHandler());
    File experimentFile = 
        new File(xcheckHome + experimentFolder + "experiment.xml");
    Document experiment = null;
    try {
      experiment = parserExperiment.parse(experimentFile);
    } catch (IOException ioe) {
      printErrorMessage(
          "There was a problem while parsing the experiment document. I stop here",
                       ioe.toString());
      System.exit(1);
    } catch (SAXException se) {
      printErrorMessage(
          "There was a problem while parsing the experiment document. I stop here",
                        se.toString());
      System.exit(1);
    } 
    
    // Validate the experiment document wrt a given W3C XML Schema
    /*
    SchemaFactory factory = 
    SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
    Schema schema = 
    factory.newSchema(new File(xcheckHome + "schema" + SEP + "experiment.xsd"));
    Validator validator = schema.newValidator();
    try {
      validator.validate(new DOMSource(experiment));
    } catch (SAXException e) {
      System.err.println(e);
      System.exit(1);
    }
    */
    
    // Query the experiment document
    xpath = XPathFactory.newInstance().newXPath();
    
    try {
      // Retrieve experiment name
      String searchName = "/experiment/name";
      experimentName = xpath.evaluate(searchName, experiment);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
      
    try {  
      // Retrieve experiment description
      String searchDesc = "/experiment/description";
      experimentDesc = xpath.evaluate(searchDesc, experiment);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
  
    // Retrieve queries
    String searchQueries = "/experiment/queries/query";
    NodeList queries = null;
    try {
      queries = (NodeList) 
          xpath.evaluate(searchQueries, experiment, XPathConstants.NODESET);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    numQueries = queries.getLength();
    
    // Retrieve query ids
    queryID = new String[numQueries];
    for(int i = 0; i < numQueries; i++) {
      queryID[i] = 
            ((Element) queries.item(i)).getAttribute("id");
    }
    
    // Retrieve query descriptions
    queryDesc = new String[numQueries];
    for(int i = 0; i < numQueries; i++) {
      NodeList nl = ((Element) queries.item(i)).getElementsByTagName("description");
      if (nl.getLength() != 0) {
        queryDesc[i] = nl.item(0).getFirstChild().getNodeValue();
      } else {
        queryDesc[i] = "";
      }
    }
    
    // Retrieve query syntax
    querySyntax = new String[numQueries];
    for(int i = 0; i < numQueries; i++) {
      NodeList nl = ((Element) queries.item(i)).getElementsByTagName("syntax");
      if (nl.getLength() != 0) {
        querySyntax[i] = nl.item(0).getFirstChild().getNodeValue();
      } else {
        querySyntax[i] = "";
      }
    }

    
    // Retrieve query files
    queryFileName = new String[numQueries];
    for(int i = 0; i < numQueries; i++) {
      NodeList nl = ((Element) queries.item(i)).getElementsByTagName("file");
      if (nl.getLength() != 0) {
        queryFileName[i] = nl.item(0).getFirstChild().getNodeValue();
      } else {
        queryFileName[i] = "";
      }
    }
    
    // Retrieve query types
    queryType = new String[numQueries];
    for(int i = 0; i < numQueries; i++) {
      queryType[i] = ((Element) queries.item(i)).getAttribute("type");
      if (queryType[i].equals("")) {
        queryType[i] = "nodeset";
      }
    }

    // Retrieve query targets
    queryTargetID = new String[numQueries];
    queryTargetDesc = new String[numQueries];
    queryTargetFile = new String[numQueries];
    for(int i = 0; i < numQueries; i++) {
      queryTargetID[i] = 
          ((Element) queries.item(i)).getAttribute("target");
      try {  
        // Retrieve target description
        String search = "id(@target)/description";
        queryTargetDesc[i] = xpath.evaluate(search, queries.item(i));
      } catch (XPathExpressionException xee) {xee.printStackTrace();}
      try {  
        // Retrieve target file
        String search = "id(@target)/file";
        queryTargetFile[i] = xpath.evaluate(search, queries.item(i));
      } catch (XPathExpressionException xee) {xee.printStackTrace();}
    }

    
    
    
        
    /************************************
    Check for existence of outcome folder 
    *************************************/    
      // If Java version >= 1.6 use the following 'Console' code:
/*    if (checkOutcome.equals("yes")) {
        folder = new File(xcheckHome + experimentFolder + "outcome_" + engineID);
      if (folder.exists()) {
        Console console = System.console();
        if (console != null) {
          console.format("%nA folder named outcome already exists in " + 
              experimentFolder + "%n"); 
          boolean consoleAnswer = true;
          while (consoleAnswer) {
            String consoleInput = 
                console.readLine("Should I overwrite it? [yes | no]: ");
            if (consoleInput.equals("yes")) {
              consoleAnswer = false;
            }
            if (consoleInput.equals("no")) {
              System.exit(1);
            }
          }
        } else {
          System.out.println(
              "A folder named outcome already exists, " +
              "the checkOutcome option in the configuration file has value 'yes', " +
              "and no console is available. I stop here.");
          System.exit(1);
        }
      }
    }
*/
    
    
    // If Java version <= 1.5 replace the above 'Console' code with the following:
    if(checkOutcome.equals("yes")) {
      folder = new File(xcheckHome + experimentFolder + "outcome_" + engineID);
      if(folder.exists()) {
        System.out.printf("%nA folder named outcome already exists in %s%n",
        experimentFolder);
        String answer = "";
        while(!answer.equals("yes")) {
          System.out.print("Should I overwrite it? [yes | no]: ");
          try {
            final InputStreamReader isr = new InputStreamReader(System.in);
            answer = new BufferedReader(isr).readLine().trim();
            if(answer.equals("no")) System.exit(1);
          } catch(final Exception e) {
            System.exit(1);
          }
        }
      }
    } 
    
    
    /*************************************
    Remove old content from outcome folder 
    **************************************/    
    folder = new File(
        xcheckHome + experimentFolder + "outcome_" + engineID + SEP + "answers");
    if (folder.exists()) {
      File[] files = folder.listFiles();
      for (int i = 0; i < files.length; i++) {
        files[i].delete();
      }
    }
    folder = new File(xcheckHome + experimentFolder + "outcome_" + 
        engineID + SEP + "plots");
    if (folder.exists()) {
      File[] files = folder.listFiles();
      for (int i = 0; i < files.length; i++) {
        files[i].delete();
      }
    }
    folder = new File(
        xcheckHome + experimentFolder + "outcome_" + engineID + SEP + "outcome.html");
    if (folder.exists()) {
      folder.delete();
    }
    folder = new File(
        xcheckHome + experimentFolder + "outcome_" + engineID + SEP + "times.xml");
    if (folder.exists()) {
      folder.delete();
    }
    folder = new File(
        xcheckHome + experimentFolder + "outcome_" + engineID + SEP + "errorLog.xml");
    if (folder.exists()) {
      folder.delete();
    }


    /***************************************************************
    Create a new outcome folder and subfolders (if they do not exist)
    ***************************************************************/    
    folder = new File(xcheckHome + experimentFolder + "outcome_" + engineID);
    if (! folder.exists()) {
      folder.mkdir();
    }
    folder = new File(xcheckHome + experimentFolder + "outcome_" + 
        engineID + SEP + "answers");
    if (! folder.exists()) {
      folder.mkdir();
    }
    folder = new File(xcheckHome + experimentFolder + "outcome_" + 
        engineID + SEP + "plots");
    if (! folder.exists()) {
      folder.mkdir();
    }

    
    /***************************** 
    Execute the before-run command
    *****************************/    
    if (engineBeforeRun != "") {
      Process proc = null;
      try {
        proc = Runtime.getRuntime().exec(engineBeforeRun);
      } catch (IOException ioe) {
        printErrorMessage("There was a problem while executing before-run command",
                          ioe.toString());
      }
    }
    
    
    
    /***************** 
    Run the experiment
    *****************/    
    System.out.println();
    System.out.println("Executing the experiment:");
    System.out.println();
    // Parsing times
    timeVectorParse = new long[numQueries];
    // Query execution times
    timeVectorExecute = new long[numQueries];
    // Result serialization times
    timeVectorSerialize = new long[numQueries];
    // Total times (parsing + execution + serialization)
    timeVectorOverall = new long[numQueries];
    // Errors in parsing documents
    errorDocument = new boolean[numQueries];
    // Error messages in parsing documents
    errorMessageDocument = new String[numQueries];
    // Errors in parsing queries
    errorQuery = new boolean[numQueries];
    // Error messages in parsing queries
    errorMessageQuery = new String[numQueries];
    long timeStart, timeEnd;


    /************* 
    For each query
    *************/
    for(int j = 0; j < numQueries; j++) {
      if (j < (numQueries - 1))
        System.out.print(queryID[j] + " ");
      else
        System.out.println(queryID[j]);
      
      /*****************
      Parse the document
      *****************/    
      String documentFullFileName =  xcheckHome + experimentFolder + 
          "documents" + SEP + queryTargetFile[j];
      Document document = null;
      timeStart = System.currentTimeMillis();
      for (int k = 0; k < runs; k++) {
        document = parseInputDocument(j, documentFullFileName);
        if (errorDocument[j]) {break;}
      }    
      if (errorDocument[j]) {continue;}
      timeEnd = System.currentTimeMillis();
      timeVectorParse[j] = (timeEnd - timeStart) / runs;

      
      /**************** 
      Execute the query
      ****************/    
      // Read query file into a string
      if (queryFileName[j] != "") { 
        BufferedReader queryFileReader = null;
        String line = null;
        String fullQueryFileName = 
            xcheckHome + experimentFolder + "queries/" + queryFileName[j]; 
        try {
          queryFileReader = new BufferedReader(new FileReader(fullQueryFileName));
          while ((line = queryFileReader.readLine()) != null) {
            querySyntax[j] = querySyntax[j] + line + String.format("%n");
          }
        } 
        catch (IOException ioe) {ioe.printStackTrace();}
        finally {
          if (queryFileReader != null) {
            try {
              queryFileReader.close();
            } catch (IOException ioe) {ioe.printStackTrace();}
          }
        }
      }
      
      // Execute the query
      Object result = null;
      timeStart = System.currentTimeMillis();
      for (int k = 0; k < runs; k++) {
        result = executeInputQuery(j, document, querySyntax[j]);
        if (errorQuery[j]) {break;}
      }
      if (errorQuery[j]) {continue;}
      timeEnd = System.currentTimeMillis();
      timeVectorExecute[j] = (timeEnd - timeStart) / runs;

      /*************************
      Serialize the query result 
      *************************/
      if (answersOption.equals("yes")) {
        String resultFileName = xcheckHome + experimentFolder + "outcome_" + 
            engineID + SEP + "answers" + SEP + "answer" + "_" + queryID[j] + ".xml";
        timeStart = System.currentTimeMillis();
        for (int k = 0; k < runs; k++) {
          serializeResult(j, result, resultFileName, document);
        }
        timeEnd = System.currentTimeMillis();
        timeVectorSerialize[j] = (timeEnd - timeStart) / runs;
      }
    } // end of processing queries
    
    
    
    /***************************** 
    Execute the after-run command
    *****************************/    
    if (engineAfterRun != "") {
      Process proc = null;
      try {
        proc = Runtime.getRuntime().exec(engineAfterRun);
      } catch (IOException ioe) {
        printErrorMessage("There was a problem while executing after-run command",
                          ioe.toString());
      }
    }
    
    
    /******************** 
    Compute overall times
    ********************/
    for(int j = 0; j < numQueries; j++) {
      if (errorDocument[j] || errorQuery[j]) {
        timeVectorOverall[j] = 0;
      } else {
        timeVectorOverall[j] = 
          timeVectorParse[j] + timeVectorExecute[j] + timeVectorSerialize[j];
      }
    }
  
    /******************************************** 
    Print experiment's outcomes to standard output
    *********************************************/
    if (timesOption.equals("yes")) {
      printTimeToOut();
    }

    /************************************** 
    Check for errors and save the error log
    **************************************/
    errors = false;
    for(int j = 0; (! errors) && (j < numQueries); j++) {
      if (errorDocument[j] || errorQuery[j]) {
        errors = true;
      }
    }
    if (errors) {
      saveErrorLog();
    }
    
    /************************************************** 
    Generate an XML document with experiment's outcomes
    **************************************************/
    if (xmlOption.equals("yes") && timesOption.equals("yes")) {
      generateXML();
    }
    
    /************************************************ 
    Generate Gnuplot plots with experiment's outcomes
    ************************************************/
    if ((plotOption.equals("yes") || plotOption.equals("noImage")) &&
        timesOption.equals("yes")) {
      generatePlot();
    }
    
    /****************************************************
    Generate an XHTML document with experiment's outcomes
    ****************************************************/
    if (htmlOption.equals("yes")) {
      generateHTML();
    }
    
  } // end main
  
  
  
  /********************************************************************************** 
  Parse the input document. 
  Input: 
      int j: the index of the query;
      String target: the name of the file containing the document to parse;
  Output: 
      An org.w3c.dom Document object corresponding to the input document. 
  Note: If an error occurs while parsing the document, then the error message should 
      be saved in errorMessageDocument[j] and errorDocument[j] should be set to true. 
  **********************************************************************************/
  public abstract Document parseInputDocument(int j, String target); 
  
        
  /*******************************************************************************
  Execute a query. 
  Input:
    int j: the index of the query;
    Document document: the input document on which the query should run; 
    String query: the query string to execute;
  Output: 
    An Object with the query result.
  Note: 
      if an error occurs while parsing the query, then the error message should be
      saved in errorMessageQuery[j] and errorQuery[j] should be set to true.
      Moreover, the method printErrorMessage can be used to print the error on 
      standard error.
  *******************************************************************************/
  public abstract Object executeInputQuery(int j, Document document, 
                                           String query);
  
  /***************************************************************************
  Serialize the query result. 
  Input:
    int j: the index of the query;
    Object result: the query result; 
    String resultFileName: the name of the result file;
    Document document: the input document. This parameter might be useful to create 
    the serializer).
  ***************************************************************************/
  public abstract void serializeResult(int j, Object result, String resultFileName, 
                                       Document document);
  
  
  /********************** 
  Save the error log file 
  **********************/
  public void saveErrorLog() {
        
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder parser = null;
    try {
      parser = factory.newDocumentBuilder();
    } catch (ParserConfigurationException pce) {pce.printStackTrace();}
    
    // Generate the error log XML document
    Document errorLog = parser.newDocument();
    Element experiment = errorLog.createElement("experiment");
    experiment.appendChild(createLeaf(errorLog, "name", experimentName));
    experiment.appendChild(createLeaf(errorLog, "description", experimentDesc));
    Element errors = errorLog.createElement("errors");
    for(int j = 0; j < numQueries; j++) {
      if (errorDocument[j] || errorQuery[j]) {
        Element query = errorLog.createElement("query");
        query.setAttribute("id", queryID[j]);
        if (errorDocument[j]) {
          query.appendChild(errorLog.createCDATASection(errorMessageDocument[j]));
        } else {
          query.appendChild(errorLog.createCDATASection(errorMessageQuery[j]));
        }
        errors.appendChild(query);
      }
    }
    experiment.appendChild(errors);
    errorLog.appendChild(experiment);
    
    // Serialize the error log XML document
    DOMImplementationLS implementation = 
        (DOMImplementationLS) errorLog.getImplementation();
    LSOutput output = implementation.createLSOutput();
    FileWriter outputStream = null;
    try {
      outputStream = new FileWriter(
          xcheckHome + experimentFolder + "outcome_" + 
          engineID + SEP + "errorLog.xml");
    } catch (IOException ioe) {ioe.printStackTrace();}
    output.setCharacterStream(outputStream);
    LSSerializer serializer = implementation.createLSSerializer();
    //serializer.getDomConfig().setParameter("format-pretty-print",true);
    serializer.write(errorLog, output);
    try {
      outputStream.close();
    } catch (IOException ioe) {ioe.printStackTrace();}
    
    
  }        

  
  /********************************************* 
  Print experiment's outcomes to standard output 
  *********************************************/
  public void printTimeToOut() {
    
    System.out.println();
    System.out.println("Times in milliseconds:");
    
    if (parseOption.equals("yes")) {
      System.out.println();
      System.out.println("DOCUMENT PARSING");
      for(int j = 0; j < numQueries; j++) {
        System.out.print(queryTargetID[j] + "\t");
      }
      System.out.println();
      for(int j = 0; j < numQueries; j++) {
        if (errorDocument[j]) {
          System.out.print("error" + "\t");
        } else {
          System.out.print(timeVectorParse[j] + "\t");
        }
      }
      System.out.println();
    }
    
    if (executeOption.equals("yes")) {
      System.out.println();
      System.out.println("QUERY PROCESSING");
      for(int j = 0; j < numQueries; j++) {
        System.out.print(queryID[j] + "\t");
      }
      System.out.println();
      for(int j = 0; j < numQueries; j++) {
        if (errorDocument[j] || errorQuery[j]) {
          System.out.print("error" + "\t");
        } else {
          System.out.print(timeVectorExecute[j] + "\t");
        }
      }
      System.out.println();
    }
      
    if (serializeOption.equals("yes")) {
      System.out.println();
      System.out.println("RESULT SERIALIZATION");
      for(int j = 0; j < numQueries; j++) {
        System.out.print(queryID[j] + "\t");
      }
      System.out.println();
      for(int j = 0; j < numQueries; j++) {
        if (errorDocument[j] || errorQuery[j]) {
          System.out.print("error" + "\t");
        } else {
          System.out.print(timeVectorSerialize[j] + "\t");
        }
      }
      System.out.println();
    }
    
    if (overallOption.equals("yes")) {
      System.out.println();
      System.out.println("OVERALL PROCESSING");
      for(int j = 0; j < numQueries; j++) {
        System.out.print(queryID[j] + "\t");
      }
      System.out.println();
      for(int j = 0; j < numQueries; j++) {
        if (errorDocument[j] || errorQuery[j]) {
          System.out.print("error" + "\t");
        } else {
          System.out.print(timeVectorOverall[j] + "\t");
        }
      }
      System.out.println();
    }
  }

  /************************************************** 
  Generate an XML document with experiment's outcomes 
  **************************************************/
  public void generateXML() {
    
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder parser = null;
    try {
      parser = factory.newDocumentBuilder();
    } catch (ParserConfigurationException pce) {pce.printStackTrace();}
    
    // Create the XML document
    Document doc = parser.newDocument();
    Element experiment = doc.createElement("experiment");
    experiment.appendChild(createLeaf(doc,"name",experimentName));
    experiment.appendChild(createLeaf(doc,"description",experimentDesc));
    Element times = doc.createElement("times");
    
    Element engine = doc.createElement("engine");
    engine.setAttribute("id", engineID);
    engine.appendChild(createLeaf(doc,"name",engineName));
    if (engineVersion != "") {
      engine.appendChild(createLeaf(doc,"version",engineVersion));
    }
    if (engineDesc != "") {
      engine.appendChild(createLeaf(doc,"description",engineDesc));
    }
    if (engineHomepage != "") {
      engine.appendChild(createLeaf(doc,"homepage",engineHomepage));
    }
    experiment.appendChild(engine);

    
    for(int j = 0; j < numQueries; j++) {
      Element query = doc.createElement("query");
      query.setAttribute("id", queryID[j]);
      
      if (parseOption.equals("yes")) {
        Element parse = doc.createElement("documentParse");
        if (errorDocument[j]) {
          parse.appendChild(doc.createTextNode("error"));
        } else {
          parse.appendChild(doc.createTextNode(String.valueOf(timeVectorParse[j])));
        }
        query.appendChild(parse);
      }

      
      if (executeOption.equals("yes")) {
        Element execute = doc.createElement("execute");
        if (errorDocument[j] || errorQuery[j]) {
          execute.appendChild(doc.createTextNode("error"));
        } else {
          execute.appendChild(doc.createTextNode(
              String.valueOf(timeVectorExecute[j])));
        }
        query.appendChild(execute);
      }
      
      if (serializeOption.equals("yes")) {
        Element serialize = doc.createElement("serialize");
        if (errorDocument[j] || errorQuery[j]) {
          serialize.appendChild(doc.createTextNode("error"));
        } else {
          serialize.appendChild(doc.createTextNode(
              String.valueOf(timeVectorSerialize[j])));
        }
        query.appendChild(serialize);
      }
      
      if (overallOption.equals("yes")) {
        Element total = doc.createElement("overall");
        if (errorDocument[j] || errorQuery[j]) {
          total.appendChild(doc.createTextNode("error"));
        } else {
          total.appendChild(doc.createTextNode(
              String.valueOf(timeVectorOverall[j])));
        }
        query.appendChild(total);
      }
      times.appendChild(query);
    }
    experiment.appendChild(times);
    doc.appendChild(experiment);
    
    // Serialize the XML document
    DOMImplementationLS implementation = 
        (DOMImplementationLS) doc.getImplementation();
    LSOutput output = implementation.createLSOutput();
    FileWriter outputStream = null;
    try {
      outputStream = new FileWriter(
          xcheckHome + experimentFolder + "outcome_" + engineID + SEP + "times.xml");
    } catch (IOException ioe) {ioe.printStackTrace();}
    output.setCharacterStream(outputStream);
    LSSerializer serializer = implementation.createLSSerializer();
    //serializer.getDomConfig().setParameter("format-pretty-print",true);
    serializer.write(doc, output);
    try {
      outputStream.close();
    } catch (IOException ioe) {ioe.printStackTrace();}
    
    
  }
  
  /************************************************ 
  Generate Gnuplot plots with experiment's outcomes 
  ************************************************/
  public void generatePlot() {
 
    /************************* 
    Generate DAT Gnuplot files
    *************************/    
    
    if (executeOption.equals("yes")) {
      PrintWriter execute = null;
      try {
        execute = new PrintWriter(xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "execute.dat");
      } catch (IOException ioe) {ioe.printStackTrace();}
      for(int j = 0; j < numQueries; j++) {
        execute.println((j+1) + "\t" + timeVectorExecute[j]);
      }
      execute.close();
    }

    if (serializeOption.equals("yes")) {
      PrintWriter serialize = null;
      try {
        serialize = new PrintWriter(xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "serialize.dat");
      } catch (IOException ioe) {ioe.printStackTrace();}
      for(int j = 0; j < numQueries; j++) {
        serialize.println((j+1) + "\t" + timeVectorSerialize[j]);
      }
      serialize.close();
    }

    if (overallOption.equals("yes")) {
      PrintWriter overall = null;
      try {
        overall = new PrintWriter(xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "overall.dat");
      } catch (IOException ioe) {ioe.printStackTrace();}
      for(int j = 0; j < numQueries; j++) {
        overall.println((j+1) + "\t" + timeVectorOverall[j]);
      }
      overall.close();
    }

    
    /************************* 
    Generate GPL Gnuplot files
    *************************/    
    if (executeOption.equals("yes")) {
      PrintWriter execute = null;
      try {
        execute = new PrintWriter(xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "execute.gpl");
      } catch (FileNotFoundException fnfe) {fnfe.printStackTrace();}
      generateGPL("execute", execute);
      execute.close();
    }
    
    if (serializeOption.equals("yes")) {
      PrintWriter serialize = null;
      try {
        serialize = new PrintWriter(xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "serialize.gpl");
      } catch (FileNotFoundException fnfe) {fnfe.printStackTrace();}
      generateGPL("serialize", serialize);
      serialize.close();
    }
    
    if (overallOption.equals("yes")) {
      PrintWriter overall = null; 
      try {
        overall = new PrintWriter(xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "overall.gpl");
      } catch (FileNotFoundException fnfe) {fnfe.printStackTrace();}
      generateGPL("overall", overall);
      overall.close();
    }
    
    /************************* 
    Generate PNG Gnuplot plots
    *************************/    
    if (plotOption.equals("yes")) {
    
      if (executeOption.equals("yes")) {
        Process proc = null;
        try {
          proc = Runtime.getRuntime().exec(
            gnuplotPath + "gnuplot " + xcheckHome + experimentFolder + 
            "outcome_" + engineID + SEP + "plots" + SEP + "execute.gpl");
        } catch (IOException ioe) {
          printErrorMessage(
              "There was a problem while generating plot " +  
              "execute.gpl with Gnuplot.", 
              ioe.toString());
        }
        if (proc != null) { // the process exists
          int exit = 0;
          try {
            exit = proc.waitFor();
            if (exit != 0) {
              printErrorMessage("There was a problem while generating plot " + 
                  "execute.gpl with Gnuplot.",
              getErrorMessage(proc.getErrorStream()));
            }
          } 
          catch (InterruptedException ie) {ie.printStackTrace();}
        }
      }
  
      if (serializeOption.equals("yes")) {
        Process proc = null;
        try {
          proc = Runtime.getRuntime().exec(
              gnuplotPath + "gnuplot " + xcheckHome + experimentFolder +
              "outcome_" + engineID + SEP + "plots" + SEP + "serialize.gpl");
        } catch (IOException ioe) {
          printErrorMessage("There was a problem while generating plot " +  
              "serialize.gpl with Gnuplot.", ioe.toString());
        }
        if (proc != null) { // the process exists
          int exit = 0;
          try {
            exit = proc.waitFor();
            if (exit != 0) {
              printErrorMessage("There was a problem while generating plot " +  
                  "serialize.gpl with Gnuplot.", 
              getErrorMessage(proc.getErrorStream()));
            }
          } 
          catch (InterruptedException ie) {ie.printStackTrace();}
        }
      }
      
      if (overallOption.equals("yes")) {
        Process proc = null;
        try {
          proc = Runtime.getRuntime().exec(
              gnuplotPath + "gnuplot " + xcheckHome + experimentFolder +
              "outcome_" + engineID + SEP + "plots" + SEP + "overall.gpl");
        } catch (IOException ioe) {
          printErrorMessage(
              "There was a problem while generating plot " +  
              "overall.gpl with Gnuplot.",
              ioe.toString());
        }
        if (proc != null) { // the process exists
          int exit = 0;
          try {
            exit = proc.waitFor();
            if (exit != 0) {
              printErrorMessage("There was a problem while generating plot " +  
                  "overall.gpl with Gnuplot.", 
              getErrorMessage(proc.getErrorStream()));
            }
          } 
          catch (InterruptedException ie) {ie.printStackTrace();}
        }
      }
    }
  } 
      
  
  /**************************************************** 
  Generate an XHTML document with experiment's outcomes 
  ****************************************************/
  public void generateHTML() {
    
    /***************************************** 
    Parse the outcome template (an XTHML file)
    *****************************************/
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder parser = null;
    try {
      parser = factory.newDocumentBuilder();
    } catch (ParserConfigurationException pce) {pce.printStackTrace();}
    File outcomeModel = 
        new File(xcheckHome + "schema" + SEP + "outcomeTemplate.html");
    Document outcome = null;
    try {
      outcome = parser.parse(outcomeModel);
    } 
    catch (IOException ioe) {ioe.printStackTrace();}
    catch (SAXException se) {se.printStackTrace();}

    
    /************************************************************* 
    Retrieve and fill-up relevant elements of the outcome template
    *************************************************************/
    Element div = null;
    Element a = null;
    Element img = null;
    Element table = null;
    Element tbody = null;
    Element tr = null;
    Element td = null;
    Element th = null;
    Element acronym = null;

    XPath xpath = XPathFactory.newInstance().newXPath();
    String searchTitle = "/html/head/title";
    Node title = null;
    try {
      title = (Node) xpath.evaluate(searchTitle, outcome, XPathConstants.NODE);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    title.appendChild(outcome.createTextNode("Outcome for experiment " +
        experimentName));
    
    String searchPage = "/html/body/div[@class=\"page\"]";
    Node page =  null;
    try {
      page = (Node) xpath.evaluate(searchPage, outcome, XPathConstants.NODE);
    } catch (XPathExpressionException xee) {xee.printStackTrace();}
    
    // Add header
    page.appendChild(
        createLeaf(outcome, "h1","Outcome for experiment " + experimentName));
    
    /****************************
    Add description and error log
    ****************************/
    Node par = outcome.createElement("p");
    par.appendChild(outcome.createTextNode(experimentDesc + " (see "));
    a = createLeaf(outcome, "a", "experiment document");
    a.setAttribute("href",".." + SEP + "experiment.xml");
    par.appendChild(a);   
    par.appendChild(outcome.createTextNode(" for details) processed with "));
    if (engineHomepage != "") {
      a = createLeaf(outcome, "a", engineName);
      a.setAttribute("href",engineHomepage);
      par.appendChild(a);   
    } else {
      par.appendChild(outcome.createTextNode(engineName));
    }
    if (engineVersion != "") {
      par.appendChild(outcome.createTextNode(" version " + engineVersion));
    }
    par.appendChild(outcome.createTextNode(" (see "));
    a = createLeaf(outcome, "a", "adapter document");
    a.setAttribute("href", ".." + SEP + ".." + SEP + ".." + SEP + 
        "adapters" + SEP + adapterFileName);
    par.appendChild(a);   
    par.appendChild(outcome.createTextNode(" for details)."));
    page.appendChild(par);

    
    /*******************
    Add processing times
    *******************/
    if (timesOption.equals("yes")) {

      // Add header
      page.appendChild(createLeaf(outcome, "h2","Processing times"));
      
      // Add par
      par = outcome.createElement("p");
      if (xmlOption.equals("yes")) {
        par.appendChild(outcome.createTextNode("Processing times in "));
        a = createLeaf(outcome, "a", "XML format");
        a.setAttribute("href","times.xml");
        par.appendChild(a);
        par.appendChild(outcome.createTextNode(". "));
      }
      
      String text = "All times are elapsed real times expressed in milliseconds." + 
          " Each query ran "; 
      if (runs == 1) {
        text = text + "once.";
      }
      if (runs == 2) {
        text = text + 
            "twice and the average processing time for each execution phase has been computed.";
      }
      if (runs > 2) {
        text = text + runs + 
            " times and the average processing time for each execution phase has been computed.";
      }
      par.appendChild(outcome.createTextNode(text));
      page.appendChild(par);
      /********************
      Add table for parsing
      ********************/
      if (parseOption.equals("yes")) {
        page.appendChild(createLeaf(outcome, "h3","Document parsing"));
        table = outcome.createElement("table");
        table.setAttribute("border","1");
        table.setAttribute("cellpadding","2pt");
        tbody = outcome.createElement("tbody");
        tr = outcome.createElement("tr");
        for(int i = 0; i < numQueries; i++) {
          acronym = createLeaf(outcome, "acronym", queryTargetID[i]);
          acronym.setAttribute("title", queryTargetDesc[i]);
          th = outcome.createElement("th");
          th.appendChild(acronym);
          tr.appendChild(th);
        }
        tbody.appendChild(tr);
        
        tr = outcome.createElement("tr");
        for(int i = 0; i < numQueries; i++) {
          if (errorDocument[i]) {
            a = createLeaf(outcome, "a", "error");
            a.setAttribute("href","errorLog.xml");
            td = outcome.createElement("td");
            td.appendChild(a);
            tr.appendChild(td);
          }
          else {
            tr.appendChild(
                createLeaf(outcome, "td", String.valueOf(timeVectorParse[i])));
          }
        }
        tbody.appendChild(tr);
        table.appendChild(tbody);
        page.appendChild(table);
      }
      
      /****************************
      Add table for query execution
      ****************************/
      if (executeOption.equals("yes")) {
        page.appendChild(createLeaf(outcome, "h3","Query processing"));
        table = outcome.createElement("table");
        table.setAttribute("border","1");
        table.setAttribute("cellpadding","2pt");
        tbody = outcome.createElement("tbody");
        tr = outcome.createElement("tr");
        
        for(int j = 0; j < numQueries; j++) {
          if (! queryFileName[j].equals("")) {
            a = createLeaf(outcome, "a", queryID[j]);
            a.setAttribute("href", "../queries/" + queryFileName[j]);
            th = outcome.createElement("th");
            th.appendChild(a);
            tr.appendChild(th);
          } else {
            acronym = createLeaf(outcome, "acronym", queryID[j]);
            acronym.setAttribute("title", querySyntax[j]);
            th = outcome.createElement("th");
            th.appendChild(acronym);
            tr.appendChild(th);
          }
        }
        tbody.appendChild(tr);
        
        tr = outcome.createElement("tr");
        for(int j = 0; j < numQueries; j++) {
          if (errorDocument[j] || errorQuery[j]) {
            a = createLeaf(outcome, "a", "error");
            a.setAttribute("href","errorLog.xml");
            td = outcome.createElement("td");
            td.appendChild(a);
            tr.appendChild(td);
          } else {
            tr.appendChild(
                createLeaf(outcome, "td", String.valueOf(timeVectorExecute[j])));
          }
        }
        tbody.appendChild(tr);
        
        table.appendChild(tbody);
        page.appendChild(table);
        
        /***************************
        Add plot for query execution
        ***************************/
        if (plotOption.equals("yes")) {
          div = outcome.createElement("div");
          div.setAttribute("class","plot");
          img = outcome.createElement("img");
          img.setAttribute("src","plots" + SEP + "execute.png");
          img.setAttribute("alt","Execute");
          div.appendChild(img);
          page.appendChild(div);
        }
        
        // Add plot files
        if (plotOption.equals("yes") || plotOption.equals("noImage")) {
          div = outcome.createElement("div");
          div.setAttribute("class","plotFiles");
          div.appendChild(outcome.createTextNode("Gnuplot data file: "));
          a = outcome.createElement("a");
          a.setAttribute("href","plots" + SEP + "execute.dat");
          a.appendChild(outcome.createTextNode("execute.dat"));
          div.appendChild(a);
          page.appendChild(div);
          div = outcome.createElement("div");
          div.setAttribute("class","plotFiles");
          div.appendChild(outcome.createTextNode("Gnuplot plot file: "));
          a = outcome.createElement("a");
          a.setAttribute("href","plots" + SEP + "execute.gpl");
          a.appendChild(outcome.createTextNode("execute.gpl"));
          div.appendChild(a);
          page.appendChild(div);
        }
      }
  
      /*********************************
      Add table for result serialization
      *********************************/
      if (serializeOption.equals("yes")) {
        page.appendChild(createLeaf(outcome, "h3","Result serialization"));
        table = outcome.createElement("table");
        table.setAttribute("border","1");
        table.setAttribute("cellpadding","2pt");
        tbody = outcome.createElement("tbody");
        tr = outcome.createElement("tr");
        
        for(int j = 0; j < numQueries; j++) {
          if (! queryFileName[j].equals("")) {
            a = createLeaf(outcome, "a", queryID[j]);
            a.setAttribute("href", "../queries/" + queryFileName[j]);
            th = outcome.createElement("th");
            th.appendChild(a);
            tr.appendChild(th);
          } else {
            acronym = createLeaf(outcome, "acronym", queryID[j]);
            acronym.setAttribute("title", querySyntax[j]);
            th = outcome.createElement("th");
            th.appendChild(acronym);
            tr.appendChild(th);
          }
        }
        tbody.appendChild(tr);
        
        tr = outcome.createElement("tr");
        for(int j = 0; j < numQueries; j++) {
          if (errorDocument[j] || errorQuery[j]) {
            a = createLeaf(outcome, "a", "error");
            a.setAttribute("href","errorLog.xml");
            td = outcome.createElement("td");
            td.appendChild(a);
            tr.appendChild(td);
          } else {
            tr.appendChild(
                createLeaf(outcome, "td", String.valueOf(timeVectorSerialize[j])));
          }
        }
        tbody.appendChild(tr);
        
        table.appendChild(tbody);
        page.appendChild(table);
        
        /*************************
        Add plot for serialization
        *************************/
        if (plotOption.equals("yes")) {
          div = outcome.createElement("div");
          div.setAttribute("class","plot");
          img = outcome.createElement("img");
          img.setAttribute("src","plots" + SEP + "serialize.png");
          img.setAttribute("alt","Serialize");
          div.appendChild(img);
          page.appendChild(div);
        }
        
        // Add plot files
        if (plotOption.equals("yes") || plotOption.equals("noImage")) {
          div = outcome.createElement("div");
          div.setAttribute("class","plotFiles");
          div.appendChild(outcome.createTextNode("Gnuplot data file: "));
          a = outcome.createElement("a");
          a.setAttribute("href","plots" + SEP + "serialize.dat");
          a.appendChild(outcome.createTextNode("serialize.dat"));
          div.appendChild(a);
          page.appendChild(div);
          div = outcome.createElement("div");
          div.setAttribute("class","plotFiles");
          div.appendChild(outcome.createTextNode("Gnuplot plot file: "));
          a = outcome.createElement("a");
          a.setAttribute("href","plots" + SEP + "serialize.gpl");
          a.appendChild(outcome.createTextNode("serialize.gpl"));
          div.appendChild(a);
          page.appendChild(div);
        }
      }
      
      /******************************
      Add table for overall execution
      ******************************/
      if (overallOption.equals("yes")) {
        page.appendChild(createLeaf(outcome, "h3","Overall processing"));
        table = outcome.createElement("table");
        table.setAttribute("border","1");
        table.setAttribute("cellpadding","2pt");
        tbody = outcome.createElement("tbody");
        tr = outcome.createElement("tr");
        
        for(int j = 0; j < numQueries; j++) {
          if (! queryFileName[j].equals("")) {
            a = createLeaf(outcome, "a", queryID[j]);
            a.setAttribute("href", "../queries/" + queryFileName[j]);
            th = outcome.createElement("th");
            th.appendChild(a);
            tr.appendChild(th);
          } else {
            acronym = createLeaf(outcome, "acronym", queryID[j]);
            acronym.setAttribute("title", querySyntax[j]);
            th = outcome.createElement("th");
            th.appendChild(acronym);
            tr.appendChild(th);
          }
        }
        tbody.appendChild(tr);
        
        tr = outcome.createElement("tr");
        for(int j = 0; j < numQueries; j++) {
          if (errorDocument[j] || errorQuery[j]) {
            a = createLeaf(outcome, "a", "error");
            a.setAttribute("href","errorLog.xml");
            td = outcome.createElement("td");
            td.appendChild(a);
            tr.appendChild(td);
          } else {
            tr.appendChild(
                createLeaf(outcome, "td", String.valueOf(timeVectorOverall[j])));
          }
        }
        tbody.appendChild(tr);
        
        table.appendChild(tbody);
        page.appendChild(table);
        
        /*******************
        Add plot for overall
        ********************/
        if (plotOption.equals("yes")) {
          div = outcome.createElement("div");
          div.setAttribute("class","plot");
          img = outcome.createElement("img");
          img.setAttribute("src","plots" + SEP + "overall.png");
          img.setAttribute("alt","Overall");
          div.appendChild(img);
          page.appendChild(div);
        }
        
        // Add plot files
        if (plotOption.equals("yes") || plotOption.equals("noImage")) {
          div = outcome.createElement("div");
          div.setAttribute("class","plotFiles");
          div.appendChild(outcome.createTextNode("Gnuplot data file: "));
          a = outcome.createElement("a");
          a.setAttribute("href","plots" + SEP + "overall.dat");
          a.appendChild(outcome.createTextNode("overall.dat"));
          div.appendChild(a);
          page.appendChild(div);
          div = outcome.createElement("div");
          div.setAttribute("class","plotFiles");
          div.appendChild(outcome.createTextNode("Gnuplot plot file: "));
          a = outcome.createElement("a");
          a.setAttribute("href","plots" + SEP + "overall.gpl");
          a.appendChild(outcome.createTextNode("overall.gpl"));
          div.appendChild(a);
          page.appendChild(div);
        }
      }
    } // end if

    if (answersOption.equals("yes")) {
    
      // Add header
      page.appendChild(createLeaf(outcome, "h2","Query answers"));
      
      /********************
      Add table for answers
      *********************/
      table = outcome.createElement("table");
      table.setAttribute("border","1");
      table.setAttribute("cellpadding","2pt");
      tbody = outcome.createElement("tbody");
      tr = outcome.createElement("tr");
      
      for(int j = 0; j < numQueries; j++) {
        if (! queryFileName[j].equals("")) {
          a = createLeaf(outcome, "a", queryID[j]);
          a.setAttribute("href", "../queries/" + queryFileName[j]);
          th = outcome.createElement("th");
          th.appendChild(a);
          tr.appendChild(th);
        } else {
          acronym = createLeaf(outcome, "acronym", queryID[j]);
          acronym.setAttribute("title", querySyntax[j]);
          th = outcome.createElement("th");
          th.appendChild(acronym);
          tr.appendChild(th);
        }
      }
      tbody.appendChild(tr);
        
      tr = outcome.createElement("tr");
      for(int j = 0; j < numQueries; j++) {
        if (errorDocument[j] || errorQuery[j]) {
          a = createLeaf(outcome, "a", "error");
          a.setAttribute("href","errorLog.xml");
          td = outcome.createElement("td");
          td.appendChild(a);
          tr.appendChild(td);
        } else {
          a = createLeaf(outcome, "a", "XML");
          a.setAttribute("href", 
                         "answers" + SEP + "answer" + "_" + queryID[j] + ".xml");
          td = outcome.createElement("td");
          td.appendChild(a);
          tr.appendChild(td);
        }
      }
      tbody.appendChild(tr);
        
      table.appendChild(tbody);
      page.appendChild(table);
      
    }
    
    /*************************************************************** 
    Serialize the XHTML document obtained from the outcome templeate
    ***************************************************************/
    DOMImplementationLS implementation = 
        (DOMImplementationLS) outcome.getImplementation();
    LSOutput output = implementation.createLSOutput();
    FileWriter outputStream = null;
    try {
      outputStream = new FileWriter(
          xcheckHome + experimentFolder + "outcome_" + engineID 
          + SEP + "outcome.html");
    } catch (IOException ioe) {ioe.printStackTrace();}
    output.setCharacterStream(outputStream);
    LSSerializer serializer = implementation.createLSSerializer();
    // Comment next line if using Java 1.5
    //serializer.getDomConfig().setParameter("format-pretty-print",true);
    serializer.write(outcome, output);
    try {
      outputStream.close();
    } catch (IOException ioe) {ioe.printStackTrace();}
    
  }
  
  /************************* 
  Generate Gnuplot GPL files 
  *************************/
  public void generateGPL(String phase, PrintWriter phaseWriter) {
    
    phaseWriter.println("set terminal png color");
    phaseWriter.println("set output '" + 
        xcheckHome + experimentFolder + "outcome_" + engineID + 
        SEP + "plots" + SEP + phase + ".png'");
    if (phase.equals("execute")) {
      phaseWriter.println("set title 'query processing'");
    }
    if (phase.equals("serialize")) {
      phaseWriter.println("set title 'result serialization'");
    }
    if (phase.equals("overall")) {
      phaseWriter.println("set title 'overall processing'");
    }
    phaseWriter.println("set xlabel 'Query'");
    phaseWriter.println("set ylabel 'Time'");
    int _numQueries = numQueries + 1;
    phaseWriter.println("set xrange [0:" + _numQueries + "]");
    phaseWriter.println("set yrange [0:]");
    phaseWriter.println("set style data boxes");
    phaseWriter.println("set style fill solid 0.5 border -1");
    phaseWriter.println("set boxwidth 0.2");
    phaseWriter.print("set xtics (");
    for(int j = 0; j < numQueries; j++) {
      phaseWriter.print("\"" + queryID[j] + "\" " + (j+1));
      if (j < (numQueries - 1)) 
        phaseWriter.print(", ");
      else
        phaseWriter.println(")");
    }
    phaseWriter.print("plot \"" + xcheckHome + experimentFolder + "outcome_" +
        engineID + SEP + "plots" + SEP + phase + ".dat\" using 1:2 notitle");

  }
  

  /************************ 
  Create a leaf XML element 
  ************************/
  public Element createLeaf(Document doc, String tag, String text) {
    
    Element e = doc.createElement(tag);
    e.appendChild(doc.createTextNode(text));
    return e;

  }
  
  /****************************************
  Get the error message from an errorStream
  ****************************************/
  public String getErrorMessage(InputStream errorStream) {
    
    int x;
    int i = 0;
    char[] msg = new char[2000];
    try {
      while ((x = errorStream.read()) != -1) {
        if (i < msg.length) {
          msg[i++] = (char) x;
        }
      }
    } catch (IOException ioe) {ioe.printStackTrace();}
    return String.valueOf(msg, 0, i);

  }
  
  
  /****************** 
  Print error message
  *******************/
  public void printErrorMessage(String msg) {
  
    final String separator = 
"-------------------------------------------------------------------------------";
    System.err.println();
    System.err.println();
    System.err.println(separator);
    System.err.println(msg);
    System.err.println(separator);
    System.err.println();
  
  }
  
  
  /****************** 
  Print error message
  *******************/
  public void printErrorMessage(String msg1, String msg2) {
  
    final String separator = 
"-------------------------------------------------------------------------------";
    System.err.println();
    System.err.println();
    System.err.println(separator);
    System.err.println(msg1);
    System.err.println("This is what I know:");
    System.err.println(msg2);
    System.err.println(separator);
    System.err.println();
  
  }
  
}
 

/*************************************************** 
Strict error handler. It exits when an error occurs.
***************************************************/
class StrictErrorHandler extends DefaultHandler {
  
  public void printErrorMessage(String msg1, String msg2) {
  
    final String separator = 
        "-------------------------------------------------------------------------------";
    System.err.println();
    System.err.println();
    System.err.println(separator);
    System.err.println(msg1);
    System.err.println("This is what I know:");
    System.err.println(msg2);
    System.err.println(separator);
    System.err.println();
  
  }
  
  
  public void warning(SAXParseException spe) throws SAXException {
    
    printErrorMessage(
        "There was a warning message while parsing document " +
        spe.getSystemId(),
        spe.getLineNumber() + ":" + spe.getColumnNumber() + ": " + spe.toString());
  }
  
  public void error(SAXParseException spe) throws SAXException {
    
    printErrorMessage(
        "There was a problem while parsing document " +
        spe.getSystemId() + " I stop here.",
        spe.getLineNumber() + ":" + spe.getColumnNumber() + ": " + spe.toString());
    System.exit(1);
  }
  
  public void fatalError(SAXParseException spe) throws SAXException {
    
    printErrorMessage(
        "There was a problem while parsing document " +
        spe.getSystemId() + " I stop here.",
        spe.getLineNumber() + ":" + spe.getColumnNumber() + ": " + spe.toString());
    System.exit(1);
  }
  
}

/********************************** 
Loose error handler. It never exits
**********************************/
class LooseErrorHandler extends DefaultHandler {
  
  public void printErrorMessage(String msg1, String msg2) {
  
    final String separator = 
        "-------------------------------------------------------------------------------";
    System.err.println();
    System.err.println();
    System.err.println(separator);
    System.err.println(msg1);
    System.err.println("This is what I know:");
    System.err.println(msg2);
    System.err.println(separator);
    System.err.println();
  
  }
  
  public void warning(SAXParseException spe) throws SAXException {
    
    printErrorMessage(
        "There was a warning message while parsing document " +
        spe.getSystemId(),
        spe.getLineNumber() + ":" + spe.getColumnNumber() + ": " + spe.toString());
  }
  
  public void error(SAXParseException spe) throws SAXException {
    
    printErrorMessage(
        "There was a problem while parsing document " +
        spe.getSystemId(),
        spe.getLineNumber() + ":" + spe.getColumnNumber() + ": " + spe.toString());
  }
  
  public void fatalError(SAXParseException spe) throws SAXException {
    
    printErrorMessage(
        "There was a problem while parsing document " +
        spe.getSystemId(),
        spe.getLineNumber() + ":" + spe.getColumnNumber() + ": " + spe.toString());
  }
  
}